--[[
    Author: Alternator (Massiner of Nathrezim)
    Date:	2014

]]

local AddonName, AddonTable = ...;
local Engine = AddonTable.ButtonEngine;
local ScriptsItem = Engine.ScriptsItem;

local Core = Engine.Core;
local Util = Engine.Util;
local C = Engine.Constants;
local Events = Engine.Events


--[[------------------------------------------------
    UpdateAction
--------------------------------------------------]]
function ScriptsItem.UpdateAction(BFButton, Type, previousHyperlink, newHyperlink)
    if (Type and Type ~= "item") then
        return;
    end

    local Update = false;
    local ItemName, BaseItemLink = GetItemInfo(BFButton.ItemID);

    if (not BFButton.ItemLink or strmatch(BFButton.ItemLink, ":(%d+)") ~= ItemID) then
        -- Either no link recorded, or link is not for the correct item
        BFButton.ItemLink = BaseItemLink;
        Update = true;
    end

    if (newHyperlink and BFButton.ItemLink == previousHyperlink) then
        -- update action based on change to item - this item has an outdated hyperlink
        BFButton.ItemLink = newHyperlink;
        Update = true;
    end

    if (not ItemName) then
        -- we had a miss on getting item info so wait and try again
        Events.WatchForItemInfoEvent(BFButton.ItemID);
    elseif (BFButton.ItemName ~= ItemName) then
        -- The item name is not correct so update it
        BFButton.ItemName = ItemName;
        Update = true;
    end
     
    if (Update) then
        ScriptsItem.UpdateAttributes(BFButton);
        ScriptsItem.UpdateSecureCursor(BFButton);
        ScriptsItem.UpdateActionTable(BFButton);
        Core.ReportEvent(BFButton, C.EVENT_UPDATEACTION, BFButton.Action);
    end
end

function ScriptsItem.UpdateAttributes(BFButton)
    BFButton.ABW:SetAttribute("type", "item");
    BFButton.ABW:SetAttribute("item", BFButton.ItemName);
end
function ScriptsItem.UpdateSecureCursor(BFButton)
    BFButton.ABW:SetAttribute("CursorKind", "item");
    BFButton.ABW:SetAttribute("CursorTarget", BFButton.ItemID);
end
function ScriptsItem.UpdateActionTable(BFButton)
    local Action = BFButton.Action;
    Util.ClearTable(Action);
    Action.Type = "item";
    Action.ItemID = BFButton.ItemID;
    Action.ItemName = BFButton.ItemName;
    Action.ItemLink = BFButton.ItemLink;
end


--[[------------------------------------------------
    Full Update
--------------------------------------------------]]
function ScriptsItem.FullUpdate(BFButton)
    BFButton.IconTexture = GetItemIcon(BFButton.ItemID) or C.QUESTION_MARK;				--safe no matter what
    BFButton.ABWIcon:SetTexture(BFButton.IconTexture);
    BFButton:UpdateChecked();
    BFButton:UpdateEquipped();
    BFButton:UpdateUsable();
    BFButton:UpdateCooldown();
    BFButton:UpdateText();
    BFButton:UpdateRangeRegistration();
    
    if (GetMouseFocus() == BFButton.ABW) then
        BFButton.ABW:UpdateTooltip();
    end
end


--[[------------------------------------------------
    Checked
--------------------------------------------------]]
function ScriptsItem.UpdateChecked(BFButton)
    if (IsCurrentItem(BFButton.ItemID)) then
        BFButton.ABW:SetChecked(true);
    else
        BFButton.ABW:SetChecked(false);
    end
end


--[[------------------------------------------------
    Equipped
--------------------------------------------------]]
function ScriptsItem.UpdateEquipped(BFButton)
    if (IsEquippedItem(BFButton.ItemID)) then
        BFButton.ABWBorder:SetVertexColor(0, 1.0, 0, 0.5);
        BFButton.ABWBorder:Show();
    else
        BFButton.ABWBorder:Hide();
    end
end


--[[------------------------------------------------
    Usable
--------------------------------------------------]]
function ScriptsItem.UpdateUsable(BFButton)
    if (BFButton.IsToy) then
        Core.UpdateButtonUsable(BFButton, C_ToyBox.IsToyUsable(BFButton.ItemID));
    elseif (IsEquippableItem(BFButton.ItemID)) then
        Core.UpdateButtonUsable(BFButton, true);
    else
        Core.UpdateButtonUsable(BFButton, IsUsableItem(BFButton.ItemID));
    end
end


--[[------------------------------------------------
    Cooldown
--------------------------------------------------]]
function ScriptsItem.UpdateCooldown(BFButton)
    CooldownFrame_Set(BFButton.ABWCooldown, GetItemCooldown(BFButton.ItemID));
end


--[[------------------------------------------------
    Text	(Charges, or Counts)
--------------------------------------------------]]
function ScriptsItem.UpdateText(BFButton)
    local Count = GetItemCount(BFButton.ItemID, false, true);
    if (Count > 1 or IsConsumableItem(BFButton.ItemID)) then
        BFButton.ABWCount:SetText(Count);
        return;
    end
    BFButton.ABWCount:SetText(nil);
end


--[[------------------------------------------------
    Tooltip
--------------------------------------------------]]
function ScriptsItem.UpdateTooltip(ABW)
    if (ABW.BFButton.IsToy) then
        GameTooltip_SetDefaultAnchor(GameTooltip, ABW);
        GameTooltip:SetToyByItemID(ABW.BFButton.ItemID);
        return;
    end

    local InventorySlot = Util.LookupItemInventorySlot(ABW.BFButton.ItemID);
    if (InventorySlot) then
        GameTooltip_SetDefaultAnchor(GameTooltip, ABW);
        GameTooltip:SetInventoryItem("player", InventorySlot);
    else
        local Bag, Slot = Util.LookupItemBagSlot(ABW.BFButton.ItemID);
        if (Bag) then
            GameTooltip_SetDefaultAnchor(GameTooltip, ABW);
            GameTooltip:SetBagItem(Bag, Slot);
        elseif (ABW.BFButton.ItemLink) then
            GameTooltip_SetDefaultAnchor(GameTooltip, ABW);	--It appears that the sethyperlink (specifically this one) requires that the anchor be constantly refreshed!?
            GameTooltip:SetHyperlink(ABW.BFButton.ItemLink);
        end
    end
end


--[[------------------------------------------------
    Range Registration
--------------------------------------------------]]
function ScriptsItem.UpdateRangeRegistration(BFButton)
    Core.UpdateButtonRangeRegistration(BFButton, IsItemInRange(BFButton.ItemID, BFButton.Target));
end

    
--[[------------------------------------------------
    Range Check
--------------------------------------------------]]
function ScriptsItem.CheckRange(BFButton)
    Core.UpdateButtonRangeIndicator(BFButton, IsItemInRange(BFButton.ItemID, BFButton.Target));
end


--[[------------------------------------------------
    Cursor
--------------------------------------------------]]
function ScriptsItem.GetCursor(BFButton)
    return "item", BFButton.ItemID, nil, nil;
end

